#include "NMSRotated.h"
#include "Utils_eigne.h"

template<typename T>
class PostProcessFcos3D{

  public:
    PredOut<T> predict(ClsScores<T>& cls_scores,
                       BboxPreds<T>& bbox_preds,
                       DirClsPreds<T>& dir_cls_preds,
                       AttrPreds<T>& attr_preds,
                       Centerness<T>& centerness);

private:
    Eigen::MatrixXd get_points(std::array<Eigen::Index, 2> featmap_size, int stride);
    Matrix_R<T> points_img2cam(Matrix_R<T> pointsMat, Vec2D<T> view);
    Matrix_R<T> decode_yaw(Matrix_R<T> mlvl_bboxes_s, Matrix_R<T> mlvl_centers_2d_s, Eigen::VectorXi mlvl_dir_scores_s);
    Eigen::VectorX<T> limit_period_np(Eigen::VectorX<T> val, float offset, float period);
    Matrix_R<T> camera_instance_3dbox_bev(Matrix_R<T> input_array);
    Matrix_R<T> camera_instance_3dbox(Matrix_R<T> input_array);
    Vec<int> strides={8, 16, 32, 64, 128};
    int nms_pre=1000;
    bool use_direction_classifier = true;
    bool pred_attrs = true;
    int cls_out_channels=10;
    int num_attrs=9;
    Vec<int> group_reg_dims={2, 1, 3, 1, 2};
    int bbox_code_size=9;
    float dir_offset=0.7854;// pi/4
    float cfg_score_thr=0.2;
    int cfg_max_per_img=200;
    bool cfg_use_rotate_nms = true;
    float cfg_nms_thr=0.8;
    Vec<float> scale_factor={1.0, 1.0};// # test config mmdet.Resize
    // info[‘images’]: A dict contains six keys corresponding to each camera:
    // 'CAM_FRONT', 'CAM_FRONT_RIGHT', 'CAM_FRONT_LEFT', 'CAM_BACK', 'CAM_BACK_LEFT', 'CAM_BACK_RIGHT'.
    // Each dict contains all data information related to corresponding camera
    Vec2D<T> nuscenes_cam_back_cam2img = {{809.221, 0.0, 829.2196},
                                             {0.0, 809.221, 481.7784},
                                             {0.0, 0.0, 1.0}};
    Vec<float> origin = {0.5, 0.5, 0.5};
    Vec<float> origin_defualt = {0.5, 1.0, 0.5};
};

template<typename T>
PredOut<T> PostProcessFcos3D<T>::predict(
        ClsScores<T>& cls_scores,
        BboxPreds<T>& bbox_preds,
        DirClsPreds<T>& dir_cls_preds,
        AttrPreds<T>& attr_preds,
        Centerness<T>& centerness)
{
    Vec<Matrix_R<T>> mlvl_centers_2d(cls_scores.len);
    Vec<Matrix_R<T>> mlvl_bboxes(cls_scores.len);
    Vec<Matrix_R<T>> mlvl_scores(cls_scores.len);
    Vec<Eigen::VectorXi> mlvl_dir_scores(cls_scores.len);
    Vec<Eigen::VectorXi> mlvl_attr_scores(cls_scores.len);
    Vec<Matrix_R<T>> mlvl_centerness(cls_scores.len);

    for(int i=0; i<cls_scores.len; i++){
        Tensor_R<T, 3> cls_score = cls_scores.scrs[i];
        Tensor_R<T, 3> bbox_pred = bbox_preds.bboxes[i];
        Tensor_R<T, 3> dir_cls_pred = dir_cls_preds.dirs[i];
        Tensor_R<T, 3> attr_pred = attr_preds.attr[i];
        Tensor_R<T, 3> centernes = centerness.centers[i];

        // clas_score
        Eigen::array<int, 3> perm_order{1, 2, 0};
        auto in_dims = cls_score.dimensions();
        std::array<Eigen::Index, 2> out_dims{in_dims[1]*in_dims[2], in_dims[0]};
        //std::cout << "-=-=-=--=-=-=-=-=-=-=-=-=-=-=-> " <<in_dims <<std::endl;
        Tensor_R<T, 2> cls_score_reshaped =
                    cls_score.shuffle(perm_order).eval()
                                .reshape(out_dims).eval();
        Eigen::Map<Matrix_R<T>> scores(cls_score_reshaped.data(), out_dims[0], out_dims[1]);
        // FIXME: possible perfromance reduction
        scores = scores.unaryExpr(std::ref(sigmoid<T>)).eval();

        // dir_cls_pred
        in_dims = dir_cls_pred.dimensions();
        out_dims[0] = in_dims[1]*in_dims[2];
        out_dims[1] = in_dims[0];
        Tensor_R<T, 2> dir_cls_pred_reshaped =
            dir_cls_pred.shuffle(perm_order).eval()
                        .reshape(out_dims).eval();
        Eigen::Map<Matrix_R<T>> dir_cls_pred_mat(dir_cls_pred_reshaped.data(), out_dims[0], out_dims[1]);
        Eigen::VectorXi dir_cls_score;
        rowwise_maxcoeff_ind(dir_cls_pred_mat, dir_cls_score);

        // attr_pred
        in_dims = attr_pred.dimensions();
        out_dims[0] = in_dims[1]*in_dims[2];
        out_dims[1] = num_attrs;
        Tensor_R<T, 2> attr_pred_reshaped =
            attr_pred.shuffle(perm_order).eval()
                     .reshape(out_dims).eval();
        Eigen::Map<Matrix_R<T>> attr_pred_mat(attr_pred_reshaped.data(), out_dims[0], out_dims[1]);
        Eigen::VectorXi attr_pred_score;
        rowwise_maxcoeff_ind(attr_pred_mat, attr_pred_score);

        // centernes
        in_dims = centernes.dimensions();
        out_dims[0] = in_dims[0]*in_dims[1]*in_dims[2];
        out_dims[1] = 1;
        Tensor_R<T, 2> centernes_reshaped =
            centernes.shuffle(perm_order).eval()
                     .reshape(out_dims).eval();
        Eigen::Map<Matrix_R<T>> centernes_matt(centernes_reshaped.data(), out_dims[0], out_dims[1]);
        centernes_matt = centernes_matt.unaryExpr(std::ref(sigmoid<T>)).eval();


        // bbox_pred
        in_dims = bbox_pred.dimensions();
        out_dims[0] = in_dims[1]*in_dims[2];
        out_dims[1] = std::accumulate(group_reg_dims.begin(), group_reg_dims.end(), 0);
        Tensor_R<T, 2> bbox_pred_reshaped =
            bbox_pred.shuffle(perm_order).eval()
                     .reshape(out_dims).eval();
        Eigen::Map<Matrix_R<T>> bbox_pred_mat(bbox_pred_reshaped.data(), out_dims[0], out_dims[1]);

        //calculate points
        std::array<Eigen::Index, 2> featmap_size_indx{in_dims[1], in_dims[2]};
        Eigen::MatrixXd points = get_points(featmap_size_indx ,strides[i]);

        Matrix_R<T> k_points, k_bbox_pred, k_scores, k_dir_cls_pred, k_centerness;
        Eigen::VectorXi k_dir_cls_score, k_attr_score;
        if(nms_pre>0 && scores.rows() > nms_pre){
            Matrix_R<T> temp_score(scores.rows(), scores.cols());
            for(int cnt=0; cnt<scores.cols(); cnt++){
                temp_score.array().col(cnt) = scores.array().col(cnt) * centernes_matt.array().col(0);
            }

            Eigen::VectorX<T> max_scores = temp_score.rowwise().maxCoeff();

            Vec<int> topk_inds = topk(max_scores, nms_pre);

            k_points = points(topk_inds, Eigen::all).cast<T>();
            k_bbox_pred = bbox_pred_mat(topk_inds, Eigen::all);
            k_scores = scores(topk_inds, Eigen::all);
            k_dir_cls_pred = dir_cls_pred_mat(topk_inds, Eigen::all);
            k_centerness = centernes_matt(topk_inds, Eigen::all);
            k_dir_cls_score = dir_cls_score(topk_inds, Eigen::all);
            k_attr_score = attr_pred_score(topk_inds, Eigen::all);
        }
        else{
            k_points = points(Eigen::all, Eigen::all).cast<T>();
            k_bbox_pred = bbox_pred_mat(Eigen::all, Eigen::all);
            k_scores = scores(Eigen::all, Eigen::all);
            k_dir_cls_pred = dir_cls_pred_mat(Eigen::all, Eigen::all);
            k_centerness = centernes_matt(Eigen::all, Eigen::all);
            k_dir_cls_score = dir_cls_score(Eigen::all, Eigen::all);
            k_attr_score = attr_pred_score(Eigen::all, Eigen::all);

        }
        std::vector<int> ind{0,1};
        k_bbox_pred(Eigen::all, ind) = k_points(Eigen::all, ind) - k_bbox_pred(Eigen::all, ind);
        k_bbox_pred(Eigen::all, ind.at(0)) = k_bbox_pred(Eigen::all, ind.at(0))/scale_factor.at(0);
        k_bbox_pred(Eigen::all, ind.at(1)) = k_bbox_pred(Eigen::all, ind.at(1))/scale_factor.at(1);

        std::vector<int> ind2{0,1,2};
        Matrix_R<T> pred_center2d = k_bbox_pred(Eigen::all, ind2);

        k_bbox_pred(Eigen::all, ind2) = points_img2cam(pred_center2d, nuscenes_cam_back_cam2img);
        //std::cout << "pred_center2d_temp: " << pred_center2d_temp <<std::endl;
        mlvl_centers_2d[i] = pred_center2d;
        mlvl_bboxes[i] = k_bbox_pred;
        mlvl_scores[i] = k_scores;
        mlvl_dir_scores[i] = k_dir_cls_score;
        mlvl_attr_scores[i] = k_attr_score;
        mlvl_centerness[i] = k_centerness;

    }

    Matrix_R<T> mlvl_centers_2d_stacked = cat(mlvl_centers_2d);
    Matrix_R<T> mlvl_bboxes_stacked     = cat(mlvl_bboxes);
    Eigen::VectorXi mlvl_dir_scores_stacked = cat(mlvl_dir_scores);

    // Create a 3x3 identity matrix
    Matrix_R<T> cam2img = Eigen::MatrixXd::Zero(4, 4).cast<T>();
    for(int cnt=0; cnt<nuscenes_cam_back_cam2img.size(); cnt++){
        for (int cnt_vec=0; cnt_vec<nuscenes_cam_back_cam2img.at(0).size(); cnt_vec++)
           cam2img(cnt, cnt_vec) = nuscenes_cam_back_cam2img.at(cnt)[cnt_vec];
    }

    mlvl_bboxes_stacked = decode_yaw(mlvl_bboxes_stacked, mlvl_centers_2d_stacked, mlvl_dir_scores_stacked);

    Matrix_R<T> mlvl_bboxes_stacked_temp = camera_instance_3dbox_bev(mlvl_bboxes_stacked);

    Matrix_R<T> mlvl_bboxes_for_nms =
        xywhr2xyxyr(mlvl_bboxes_stacked_temp);

    Matrix_R<T> mlvl_scores_temp = cat(mlvl_scores);
    int rows_temp_ms = mlvl_scores_temp.rows();
    Matrix_R<T> padding = Eigen::MatrixXd::Zero(rows_temp_ms, 1).cast<T>();

    // Concatenate matrices horizontally
    Matrix_R<T> mlvl_scores_stacked(mlvl_scores_temp.rows(), mlvl_scores_temp.cols() + 1);
    mlvl_scores_stacked << mlvl_scores_temp, padding;

    Eigen::VectorXi mlvl_attr_scores_stacked = cat(mlvl_attr_scores);
    Matrix_R<T> mlvl_centerness_stacked = cat(mlvl_centerness);
    Matrix_R<T> mlvl_nms_scores(mlvl_scores_stacked.rows(), mlvl_scores_stacked.cols());
    for(int cnt=0; cnt<mlvl_scores_stacked.cols(); cnt++){
        mlvl_nms_scores.array().col(cnt) =  mlvl_scores_stacked.array().col(cnt)* mlvl_centerness_stacked.array().col(0);
    }

    std::cout << "\tRunning NMS...\n";
    MulticlassRes<T> results =
        box3d_multiclass_nms(mlvl_bboxes_stacked, mlvl_bboxes_for_nms,
                             mlvl_nms_scores, cfg_score_thr, cfg_max_per_img,
                             mlvl_dir_scores_stacked);

    results.bboxes = camera_instance_3dbox(results.bboxes);
    //std::cout << "results: " << results.bboxes <<std::endl;
    //std::cout << "results: " << results.scores <<std::endl;
    //std::cout << "results: " << results.labels <<std::endl;

    std::cout << "\tRunning predict...\n";
    PredOut<T> res_struct;
    res_struct.bboxes = results.bboxes;
    res_struct.scores = results.scores;
    res_struct.labels = results.labels;
    return res_struct;
}

template<typename T>
Eigen::MatrixXd PostProcessFcos3D<T>::get_points(std::array<Eigen::Index, 2> featmap_size, int stride)
{

    int h = featmap_size[0];
    int w = featmap_size[1];
    // Create coordinate vectors
    Eigen::VectorXd x = Eigen::VectorXd::LinSpaced(w, 0, w-1);  // 101 points from -5 to 5
    Eigen::VectorXd y = Eigen::VectorXd::LinSpaced(h, 0, h-1);  // 101 points from -5 to 5

    // Create a 2D grid of coordinates using Eigen
    Eigen::MatrixXd X = (((x.replicate(1, y.size())).reshaped(w*h,1).eval())* (float)stride);
    X = (X.array() + floor(stride/2)).matrix();
    Eigen::MatrixXd Y = (((y.transpose().replicate(x.size(), 1)).reshaped(w*h,1).eval()) * (float)stride);
    Y = (Y.array() + floor(stride/2)).matrix();

    // Concatenate matrices horizontally
    Eigen::MatrixXd concatenatedMatrix(X.rows(), X.cols() + Y.cols());
    concatenatedMatrix << X, Y;

    // Convert the concatenated matrix to a tensor
//        Eigen::Tensor<T, 2> outputTensor;
//    Eigen::Tensor<T, 2> outputTensor = Eigen::TensorMap<Eigen::Tensor<T, 2>>(concatenatedMatrix.cast<float>().data(), concatenatedMatrix.rows(), 2);

return concatenatedMatrix;
}

template<typename T>
Matrix_R<T> PostProcessFcos3D<T>::points_img2cam(Matrix_R<T> pointsMat, Vec2D<T> view){
    if(view.size()>4 || view[0].size()>4 || pointsMat.cols()!=3){
     std::cout << "fail---->please check the size of inputs of points_img2cam funxtion"<< std::endl;
     assert(false);
    }

    std::vector<int> ind2{0,1};
    Matrix_R<T> xys = pointsMat(Eigen::all, ind2);
    Matrix_R<T> depth = pointsMat(Eigen::all, 2); //python uses reshape but i think we don't use it here
    for(int cnt=0; cnt<xys.cols(); cnt++){
        xys.array().col(cnt) *= depth.array().col(0);
    }

    Eigen::MatrixXd onesMatrix = Eigen::MatrixXd::Ones(xys.rows(), 1);
    Matrix_R<T> unnormed_xys(xys.rows(), xys.cols() + depth.cols()+1);
    unnormed_xys << xys, depth, onesMatrix.cast<T>();

    // Create a 3x3 identity matrix
    Matrix_R<T> pad_cam2img = Eigen::MatrixXd::Identity(4, 4).cast<T>();
    for(int cnt=0; cnt<view.size(); cnt++){
        for (int cnt_vec=0; cnt_vec<view.at(0).size();cnt_vec++)
           pad_cam2img(cnt, cnt_vec) = view.at(cnt)[cnt_vec];
    }

    Matrix_R<T> inv_pad_cam2img = pad_cam2img.inverse().transpose();
    std::vector<int> ind_res{0,1,2};
    Matrix_R<T> points3D = (unnormed_xys * inv_pad_cam2img)(Eigen::all, ind_res);
    return points3D;
}

template<typename T>
Matrix_R<T> PostProcessFcos3D<T>::decode_yaw(Matrix_R<T> mlvl_bboxes_s, Matrix_R<T> mlvl_centers_2d_s, Eigen::VectorXi mlvl_dir_scores_s){

    //std::cout << "rows: " << mlvl_bboxes_s.rows() <<"cols: " << mlvl_bboxes_s.cols() <<std::endl;

    Eigen::VectorX<T> dir_rot(mlvl_bboxes_s.rows(),1);
    if(mlvl_bboxes_s.rows()>0){
    dir_rot = limit_period_np(mlvl_bboxes_s(Eigen::all, 6), 0, M_PI);
    }
    //std::cout << "dir_rot: " << dir_rot <<std::endl;

    mlvl_bboxes_s(Eigen::all, 6) = dir_rot.array() + dir_offset + (M_PI*mlvl_dir_scores_s.cast<T>()).array() ;
    mlvl_bboxes_s(Eigen::all, 6) = ((mlvl_centers_2d_s(Eigen::all, 0).array() - nuscenes_cam_back_cam2img.at(0)[2]) / nuscenes_cam_back_cam2img.at(0)[0]).atan() + mlvl_bboxes_s(Eigen::all, 6).array();

    return mlvl_bboxes_s;
}

template<typename T>
Eigen::VectorX<T> PostProcessFcos3D<T>::limit_period_np(Eigen::VectorX<T> val, float offset, float period){
//val - floor(val / period + offset) * period;
 return (val.array()-dir_offset) - (floor(((val.array()-dir_offset)/period).array() + offset)*period).array();
}

template<typename T>
Matrix_R<T> PostProcessFcos3D<T>::camera_instance_3dbox_bev(Matrix_R<T> input_array){
    int cols = input_array.cols();
    int rows = input_array.rows();
    //if((input_array.rows()*input_array.cols()) == 0)

    if(cols == 6){
        Matrix_R<T> fake_rot = Eigen::MatrixXd::Zero(rows, 1).cast<T>();
        // Concatenate matrices horizontally
        Matrix_R<T> concatenatedTempMatrix(rows, cols + 1);
        concatenatedTempMatrix << input_array, fake_rot;
    }

    if(origin.at(0)!=origin_defualt.at(0) || origin.at(1)!=origin_defualt.at(1) || origin.at(2)!=origin_defualt.at(2)){
        for (int i=0; i<3; i++){
            input_array.array().col(i) += (input_array.array().col(i+3)*(origin_defualt.at(i) - origin.at(i)));
        }
    }
    std::vector<int> ind_out{0,2,3,5,6};
     Matrix_R<T> output = input_array(Eigen::all, ind_out);
     int out_cols = output.cols();
     output.array().col(out_cols-1) = (-1)*output.array().col(out_cols-1);
     return output;
}

template<typename T>
Matrix_R<T> PostProcessFcos3D<T>::camera_instance_3dbox(Matrix_R<T> input_array){
    int cols = input_array.cols();
    int rows = input_array.rows();
    //if((input_array.rows()*input_array.cols()) == 0)

    if(cols == 6){
        Matrix_R<T> fake_rot = Eigen::MatrixXd::Zero(rows, 1).cast<T>();
        // Concatenate matrices horizontally
        Matrix_R<T> concatenatedTempMatrix(rows, cols + 1);
        concatenatedTempMatrix << input_array, fake_rot;
    }

    if(origin.at(0)!=origin_defualt.at(0) || origin.at(1)!=origin_defualt.at(1) || origin.at(2)!=origin_defualt.at(2)){
        for (int i=0; i<3; i++){
            input_array.array().col(i) += (input_array.array().col(i+3)*(origin_defualt.at(i) - origin.at(i)));
        }
    }
     return input_array;
}
